'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

/** &amp; regex */
const _ar = /&/g;
/** resolve selector */
function r$(amp, selector) {
    if (selector.includes(',')) {
        return selector.split(_sr).map(part => {
            if (_ar.test(part)) {
                return part.replace(_ar, amp);
            }
            else {
                return amp + ' ' + part;
            }
        }).join(', ');
    }
    else if (_ar.test(selector)) {
        return selector.replace(_ar, amp);
    }
    else {
        return amp + ' ' + selector;
    }
}
/** seperator regex */
const _sr = /,(?![^(]*\))/;
/** trim regex */
const _tr = /\s+/g;
/** parse selector path */
function p$p(selectorPaths, instance) {
    let amp = '';
    selectorPaths.forEach(selector => {
        const adpatedSelector = selector;
        if (/,/g.test(amp)) {
            amp = amp
                .split(_sr)
                .map(ampPart => r$(ampPart, adpatedSelector))
                .join(', ');
        }
        else {
            amp = r$(amp, adpatedSelector);
        }
    });
    return amp.trim().replace(_tr, ' ');
}

/** kebab regex */
const _kr = /[A-Z\u00C0-\u00D6\u00D8-\u00DE]/g;
/** kebab case */
function _kc(pattern) {
    return pattern.replace(_kr, match => '-' + match.toLowerCase());
}
/** upwrap property */
/** TODO: refine it */
function _up(prop, indent = '  ') {
    if (typeof prop === 'object') {
        return (' {\n' +
            Object.entries(prop).map(v => {
                return indent + `  ${_kc(v[0])}: ${v[1]};`;
            }).join('\n') +
            '\n' + indent + '}');
    }
    return `: ${String(prop)};`;
}
/** unwrap properties */
function _ups(props, instance, params) {
    if (typeof props === 'function') {
        return props({
            context: instance.context,
            props: params
        });
    }
    return props;
}
/** create style */
function _cs(selector, props, instance, params) {
    if (props === null)
        return null;
    const unwrappedProps = _ups(props, instance, params);
    const propertyNames = Object.keys(unwrappedProps);
    if (propertyNames.length === 0) {
        if (instance.config.preserveEmptyBlock)
            return selector + ' {}';
        return null;
    }
    const statements = [
        selector + ' {'
    ];
    propertyNames.forEach(propertyName => {
        const property = unwrappedProps[propertyName];
        propertyName = _kc(propertyName);
        if (property !== undefined) {
            statements.push(`  ${propertyName}${_up(property)}`);
        }
    });
    statements.push('}');
    return statements.join('\n');
}
/** traverse */
function t(node, selectorPaths, styles, instance, params) {
    if (typeof node.$ === 'string') {
        selectorPaths.push(node.$);
    }
    else if (typeof node.$ === 'function') {
        selectorPaths.push(node.$({
            context: instance.context,
            props: params
        }));
    }
    else {
        if (node.$.before !== undefined)
            node.$.before(instance.context);
        if (typeof node.$.$ === 'string') {
            selectorPaths.push(node.$.$);
        }
        else {
            selectorPaths.push(node.$.$({
                context: instance.context,
                props: params
            }));
        }
    }
    const selector = p$p(selectorPaths);
    const style = _cs(selector, node.props, instance, params);
    if (style !== null)
        styles.push(style);
    if (node.children !== null) {
        node.children.forEach(childNode => {
            t(childNode, selectorPaths, styles, instance, params);
        });
    }
    selectorPaths.pop();
    if (typeof node.$ === 'object' && node.$.after !== undefined)
        node.$.after(instance.context);
}
function render(node, instance, props) {
    const styles = [];
    t(node, [], styles, instance, props);
    return styles.join('\n\n');
}

/* remove el */
function _re(el) {
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (!el)
        return;
    const parentElement = el.parentElement;
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (parentElement)
        parentElement.removeChild(el);
}
/** query element */
function _qe(id) {
    return document.querySelector(`style[css-render-id="${String(id)}"]`);
}
/** create element */
function _ce(id) {
    const el = document.createElement('style');
    el.setAttribute('css-render-id', String(id));
    return el;
}

/** get count of element */
function _gc(el) {
    const count = el.getAttribute('mount-count');
    if (count === null)
        return 0;
    return Number(count);
}
/** set count of element */
function _sc(el, count) {
    el.setAttribute('mount-count', String(count));
}
/** unmount */
function _u(intance, node, target) {
    const els = node.els;
    if (target === undefined) {
        els.forEach(_re);
        node.els = [];
    }
    else if (typeof target === 'string' || typeof target === 'number') {
        const targetElement = _qe(target);
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (targetElement && els.includes(targetElement)) {
            const mountCount = _gc(targetElement);
            if (mountCount === 1) {
                _re(targetElement);
                node.els = els.filter(el => el !== targetElement);
            }
            else {
                _sc(targetElement, mountCount - 1);
            }
        }
    }
    else if (els.includes(target)) {
        const mountCount = _gc(target);
        if (mountCount === 1) {
            _re(target);
            node.els = els.filter(el => el !== target);
        }
        else {
            _sc(target, mountCount - 1);
        }
    }
}
/** mount */
function _m(instance, node, target, props) {
    let targetElement = null;
    if (target === undefined) {
        targetElement = document.createElement('style');
        document.head.appendChild(targetElement);
    }
    else if (typeof target === 'string' || typeof target === 'number') {
        targetElement = _qe(target);
        if (targetElement === null) {
            targetElement = _ce(target);
            document.head.appendChild(targetElement);
            _sc(targetElement, 1);
        }
        else {
            _sc(targetElement, _gc(targetElement) + 1);
            return targetElement;
        }
    }
    else {
        targetElement = target;
        const mountCount = _gc(targetElement);
        if (mountCount > 0) {
            _sc(targetElement, mountCount + 1);
            return target;
        }
    }
    const style = node.render(props);
    if (targetElement.innerHTML !== style) {
        targetElement.innerHTML = style;
    }
    return targetElement;
}

/** render wrapper */
function _r(props) {
    return render(this, this.instance, props);
}
/** wrapped mount */
function _wm(options) {
    const target = options === undefined ? undefined : options.target;
    if (target === null)
        return null;
    const targetElement = _m(this.instance, this, target, options === null || options === void 0 ? void 0 : options.props);
    const els = this.els;
    if (!els.includes(targetElement)) {
        els.push(targetElement);
    }
    return targetElement;
}
/** wrapped _u */
function _wu(options) {
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    const target = !options ? undefined : options.target;
    if (target === null)
        return;
    _u(this.instance, this, target);
}
/** traverse */
function _t(children, flattenedNodes) {
    children.forEach(child => {
        if (Array.isArray(child)) {
            _t(child, flattenedNodes);
        }
        else {
            flattenedNodes.push(child);
        }
    });
}
/** flatten */
function _f(children) {
    if (children === null)
        return null;
    const flattenedNodes = [];
    _t(children, flattenedNodes);
    return flattenedNodes;
}
/** create CNode */
function _cc(instance, $, props, children) {
    return {
        instance,
        $,
        props,
        children: _f(children),
        els: [],
        render: _r,
        mount: _wm,
        unmount: _wu
    };
}
const c = function (instance, $, props, children) {
    if (Array.isArray($)) {
        return _cc(instance, '', null, $);
    }
    if (Array.isArray(props)) {
        return _cc(instance, $, null, props);
    }
    else if (Array.isArray(children)) {
        return _cc(instance, $, props, children);
    }
    else {
        return _cc(instance, $, props, null);
    }
};

function CSSRender(config = {
    preserveEmptyBlock: false
}) {
    const cssr = {
        c: ((...args) => c(cssr, ...args)),
        use: (plugin, ...args) => plugin.install(cssr, ...args),
        context: {},
        config
    };
    return cssr;
}

exports.CSSRender = CSSRender;
exports.default = CSSRender;
